@@ -6,27 +6,30 @@ Generic Python packager for Android / iOS. Desktop later.
 
 '''
 
-__version__ = '0.30'
+__version__ = '1.1.0'
 
 import os
 import re
 import sys
-import zipfile
 import select
 import codecs
+import textwrap
 from buildozer.jsonstore import JsonStore
 from sys import stdout, stderr, exit
 from re import search
 from os.path import join, exists, dirname, realpath, splitext, expanduser
 from subprocess import Popen, PIPE
-from os import environ, unlink, rename, walk, sep, listdir, makedirs
+from os import environ, unlink, walk, sep, listdir, makedirs
 from copy import copy
-from shutil import copyfile, rmtree, copytree
+from shutil import copyfile, rmtree, copytree, move
 from fnmatch import fnmatch
-try:
+
+from pprint import pformat
+
+try:  # Python 3
     from urllib.request import FancyURLopener
     from configparser import SafeConfigParser
-except ImportError:
+except ImportError:  # Python 2
     from urllib import FancyURLopener
     from ConfigParser import SafeConfigParser
 try:
@@ -40,7 +43,7 @@ try:
     colorama.init()
 
     RESET_SEQ = colorama.Fore.RESET + colorama.Style.RESET_ALL
-    COLOR_SEQ = lambda x: x
+    COLOR_SEQ = lambda x: x  # noqa: E731
     BOLD_SEQ = ''
     if sys.platform == 'win32':
         BLACK = colorama.Fore.BLACK + colorama.Style.DIM
@@ -53,7 +56,7 @@ try:
 except ImportError:
     if sys.platform != 'win32':
         RESET_SEQ = "\033[0m"
-        COLOR_SEQ = lambda x: "\033[1;{}m".format(30 + x)
+        COLOR_SEQ = lambda x: "\033[1;{}m".format(30 + x)  # noqa: E731
         BOLD_SEQ = "\033[1m"
         BLACK = 0
         RED = 1
@@ -72,11 +75,13 @@ LOG_LEVELS_T = 'EID'
 SIMPLE_HTTP_SERVER_PORT = 8000
 IS_PY3 = sys.version_info[0] >= 3
 
+
 class ChromeDownloader(FancyURLopener):
     version = (
         'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 '
         '(KHTML, like Gecko) Chrome/28.0.1500.71 Safari/537.36')
 
+
 urlretrieve = ChromeDownloader().retrieve
 
 
@@ -98,12 +103,16 @@ class BuildozerCommandException(BuildozerException):
 
 class Buildozer(object):
 
+    ERROR = 0
+    INFO = 1
+    DEBUG = 2
+
     standard_cmds = ('distclean', 'update', 'debug', 'release',
                      'deploy', 'run', 'serve')
 
     def __init__(self, filename='buildozer.spec', target=None):
         super(Buildozer, self).__init__()
-        self.log_level = 1
+        self.log_level = 2
         self.environ = {}
         self.specfilename = filename
         self.state = None
@@ -115,9 +124,13 @@ class Buildozer(object):
         self.config.getlistvalues = self._get_config_list_values
         self.config.getdefault = self._get_config_default
         self.config.getbooldefault = self._get_config_bool
+        self.config.getrawdefault = self._get_config_raw_default
 
         if exists(filename):
-            self.config.read(filename)
+            try:
+                self.config.read(filename, "utf-8")
+            except TypeError:  # python 2 has no second arg here
+                self.config.read(filename)
             self.check_configuration_tokens()
 
         # Check all section/tokens for env vars, and replace the
@@ -126,10 +139,13 @@ class Buildozer(object):
 
         try:
             self.log_level = int(self.config.getdefault(
-                'buildozer', 'log_level', '1'))
-        except:
+                'buildozer', 'log_level', '2'))
+        except Exception:
             pass
-        self.builddir = self.config.getdefault('buildozer', 'builddir', None)
+
+        self.user_bin_dir = self.config.getdefault('buildozer', 'bin_dir', None)
+        if self.user_bin_dir:
+            self.user_bin_dir = realpath(join(self.root_dir, self.user_bin_dir))
 
         self.targetname = None
         self.target = None
@@ -141,7 +157,7 @@ class Buildozer(object):
         '''
         self.targetname = target
         m = __import__('buildozer.targets.{0}'.format(target),
-                fromlist=['buildozer'])
+                       fromlist=['buildozer'])
         self.target = m.get_target(self)
         self.check_build_layout()
         self.check_configuration_tokens()
@@ -214,13 +230,19 @@ class Buildozer(object):
             print('{} {}'.format(LOG_LEVELS_T[level], msg))
 
     def debug(self, msg):
-        self.log(2, msg)
+        self.log(self.DEBUG, msg)
+
+    def log_env(self, level, env):
+        """dump env into debug logger in readable format"""
+        self.log(level, "ENVIRONMENT:")
+        for k, v in env.items():
+            self.log(level, "    {} = {}".format(k, pformat(v)))
 
     def info(self, msg):
-        self.log(1, msg)
+        self.log(self.INFO, msg)
 
     def error(self, msg):
-        self.log(0, msg)
+        self.log(self.ERROR, msg)
 
     #
     # Internal check methods
@@ -297,7 +319,7 @@ class Buildozer(object):
                     ret_stdout.append(chunk)
                 if show_output:
                     if IS_PY3:
-                        stdout.write(chunk.decode('utf-8'))
+                        stdout.write(chunk.decode('utf-8', 'replace'))
                     else:
                         stdout.write(chunk)
             if fd_stderr in readx:
@@ -308,19 +330,20 @@ class Buildozer(object):
                     ret_stderr.append(chunk)
                 if show_output:
                     if IS_PY3:
-                        stderr.write(chunk.decode('utf-8'))
+                        stderr.write(chunk.decode('utf-8', 'replace'))
                     else:
                         stderr.write(chunk)
 
-        stdout.flush()
-        stderr.flush()
+            stdout.flush()
+            stderr.flush()
 
         process.communicate()
         if process.returncode != 0 and break_on_error:
             self.error('Command failed: {0}'.format(command))
+            self.log_env(self.ERROR, kwargs['env'])
             self.error('')
             self.error('Buildozer failed to execute the last command')
-            if self.log_level <= 1:
+            if self.log_level <= self.INFO:
                 self.error('If the error is not obvious, please raise the log_level to 2')
                 self.error('and retry the latest command.')
             else:
@@ -351,7 +374,10 @@ class Buildozer(object):
         show_output = kwargs.pop('show_output')
 
         if show_output:
-            kwargs['logfile'] = codecs.getwriter('utf8')(stdout)
+            if IS_PY3:
+                kwargs['logfile'] = codecs.getwriter('utf8')(stdout.buffer)
+            else:
+                kwargs['logfile'] = codecs.getwriter('utf8')(stdout)
 
         if not sensible:
             self.debug('Run (expect) {0!r}'.format(command))
@@ -365,6 +391,7 @@ class Buildozer(object):
         '''Ensure the spec file is 'correct'.
         '''
         self.info('Check configuration tokens')
+        self.migrate_configuration_tokens()
         get = self.config.getdefault
         errors = []
         adderror = errors.append
@@ -385,13 +412,13 @@ class Buildozer(object):
             adderror('[app] One of "version" or "version.regex" must be set')
         if version and version_regex:
             adderror('[app] Conflict between "version" and "version.regex"'
-                    ', only one can be used.')
+                     ', only one can be used.')
         if version_regex and not get('app', 'version.filename', ''):
             adderror('[app] "version.filename" is missing'
-                ', required by "version.regex"')
+                     ', required by "version.regex"')
 
         orientation = get('app', 'orientation', 'landscape')
-        if orientation not in ('landscape', 'portrait', 'all'):
+        if orientation not in ('landscape', 'portrait', 'all', 'sensorLandscape'):
             adderror('[app] "orientation" have an invalid value')
 
         if errors:
@@ -401,6 +428,25 @@ class Buildozer(object):
                 print(error)
             exit(1)
 
+    def migrate_configuration_tokens(self):
+        config = self.config
+        if config.has_section("app"):
+            migration = (
+                ("android.p4a_dir", "p4a.source_dir"),
+                ("android.p4a_whitelist", "android.whitelist"),
+                ("android.bootstrap", "p4a.bootstrap"),
+                ("android.branch", "p4a.branch"),
+                ("android.p4a_whitelist_src", "android.whitelist_src"),
+                ("android.p4a_blacklist_src", "android.blacklist_src")
+            )
+            for entry_old, entry_new in migration:
+                if not config.has_option("app", entry_old):
+                    continue
+                value = config.get("app", entry_old)
+                config.set("app", entry_new, value)
+                config.remove_option("app", entry_old)
+                self.error("In section [app]: {} is deprecated, rename to {}!".format(
+                    entry_old, entry_new))
 
     def check_build_layout(self):
         '''Ensure the build (local and global) directory layout and files are
@@ -410,28 +456,26 @@ class Buildozer(object):
 
         if not exists(self.specfilename):
             print('No {0} found in the current directory. Abandon.'.format(
-                    self.specfilename))
+                self.specfilename))
             exit(1)
 
         # create global dir
         self.mkdir(self.global_buildozer_dir)
         self.mkdir(self.global_cache_dir)
 
-        # create local dir
-        specdir = dirname(self.specfilename)
-        if self.builddir:
-            specdir = self.builddir
+        # create local .buildozer/ dir
+        self.mkdir(self.buildozer_dir)
+        # create local bin/ dir
+        self.mkdir(self.bin_dir)
 
-        self.mkdir(join(specdir, '.buildozer'))
-        self.mkdir(join(specdir, 'bin'))
         self.mkdir(self.applibs_dir)
         self.state = JsonStore(join(self.buildozer_dir, 'state.db'))
 
-        if self.targetname:
-            target = self.targetname
+        target = self.targetname
+        if target:
             self.mkdir(join(self.global_platform_dir, target, 'platform'))
-            self.mkdir(join(specdir, '.buildozer', target, 'platform'))
-            self.mkdir(join(specdir, '.buildozer', target, 'app'))
+            self.mkdir(join(self.buildozer_dir, target, 'platform'))
+            self.mkdir(join(self.buildozer_dir, target, 'app'))
 
     def check_application_requirements(self):
         '''Ensure the application requirements are all available and ready to be
@@ -439,17 +483,30 @@ class Buildozer(object):
         '''
         requirements = self.config.getlist('app', 'requirements', '')
         target_available_packages = self.target.get_available_packages()
+        if target_available_packages is True:
+            # target handles all packages!
+            return
 
         # remove all the requirements that the target can compile
-        onlyname = lambda x: x.split('==')[0]
+        onlyname = lambda x: x.split('==')[0]  # noqa: E731
         requirements = [x for x in requirements if onlyname(x) not in
-                target_available_packages]
+                        target_available_packages]
+
+        if requirements and hasattr(sys, 'real_prefix'):
+            e = self.error
+            e('virtualenv is needed to install pure-Python modules, but')
+            e('virtualenv does not support nesting, and you are running')
+            e('buildozer in one. Please run buildozer outside of a')
+            e('virtualenv instead.')
+            exit(1)
 
         # did we already installed the libs ?
-        if exists(self.applibs_dir) and \
-            self.state.get('cache.applibs', '') == requirements:
-                self.debug('Application requirements already installed, pass')
-                return
+        if (
+            exists(self.applibs_dir) and
+            self.state.get('cache.applibs', '') == requirements
+        ):
+            self.debug('Application requirements already installed, pass')
+            return
 
         # recreate applibs
         self.rmdir(self.applibs_dir)
@@ -464,15 +521,10 @@ class Buildozer(object):
 
     def _install_application_requirement(self, module):
         self._ensure_virtualenv()
-        # resetup distribute, just in case
-        self.debug('Install distribute')
-        self.cmd('curl http://python-distribute.org/distribute_setup.py | venv/bin/python', get_stdout=True, cwd=self.buildozer_dir)
-
         self.debug('Install requirement {} in virtualenv'.format(module))
-        self.cmd('pip install --download-cache={} --target={} {}'.format(
-                self.global_cache_dir, self.applibs_dir, module),
-                env=self.env_venv,
-                cwd=self.buildozer_dir)
+        self.cmd('pip install --target={} {}'.format(self.applibs_dir, module),
+                 env=self.env_venv,
+                 cwd=self.buildozer_dir)
 
     def check_garden_requirements(self):
         '''Ensure required garden packages are available to be included.
@@ -558,8 +610,6 @@ class Buildozer(object):
         result = []
         for pattern in patterns:
             matches = glob(expanduser(pattern.strip()))
-            if not matches:
-                return
             result.extend(matches)
         return result
 
@@ -572,8 +622,9 @@ class Buildozer(object):
             target = join(cwd, target)
         self.debug('Rename {0} to {1}'.format(source, target))
         if not os.path.isdir(os.path.dirname(target)):
-            self.error('Rename {0} to {1} fails becaues {2} is not a directory'.format(source, target, os.path.directory(target)))
-        rename(source, target)
+            self.error(('Rename {0} to {1} fails because {2} is not a '
+                        'directory').format(source, target, target))
+        move(source, target)
 
     def file_copy(self, source, target, cwd=None):
         if cwd:
@@ -603,10 +654,7 @@ class Buildozer(object):
             return
 
         if archive.endswith('.zip'):
-            archive = join(cwd, archive)
-            zf = zipfile.ZipFile(archive)
-            zf.extractall(path=cwd)
-            zf.close()
+            self.cmd('unzip -q {}'.format(join(cwd, archive)), cwd=cwd)
             return
 
         raise Exception('Unhandled extraction for type {0}'.format(archive))
@@ -637,8 +685,9 @@ class Buildozer(object):
             else:
                 progression = '{0:.2f}%'.format(
                         index * blksize * 100. / float(size))
-            stdout.write('- Download {}\r'.format(progression))
-            stdout.flush()
+            if "CI" not in environ:
+                stdout.write('- Download {}\r'.format(progression))
+                stdout.flush()
 
         url = url + filename
         if cwd:
@@ -698,13 +747,16 @@ class Buildozer(object):
         exclude_exts = self.config.getlist('app', 'source.exclude_exts', '')
         exclude_dirs = self.config.getlist('app', 'source.exclude_dirs', '')
         exclude_patterns = self.config.getlist('app', 'source.exclude_patterns', '')
+        include_patterns = self.config.getlist('app',
+                                               'source.include_patterns',
+                                               '')
         app_dir = self.app_dir
 
         self.debug('Copy application source from {}'.format(source_dir))
 
         rmtree(self.app_dir)
 
-        for root, dirs, files in walk(source_dir):
+        for root, dirs, files in walk(source_dir, followlinks=True):
             # avoid hidden directory
             if True in [x.startswith('.') for x in root.split(sep)]:
                 continue
@@ -721,17 +773,22 @@ class Buildozer(object):
                 for exclude_dir in exclude_dirs:
                     if exclude_dir[-1] != '/':
                         exclude_dir += '/'
-                    if filtered_root.startswith(exclude_dir):
+                    if filtered_root.startswith(exclude_dir.lower()):
                         is_excluded = True
                         break
-                if is_excluded:
-                    continue
 
                 # pattern matching
-                for pattern in exclude_patterns:
+                if not is_excluded:
+                    # match pattern if not ruled out by exclude_dirs
+                    for pattern in exclude_patterns:
+                        if fnmatch(filtered_root, pattern):
+                            is_excluded = True
+                            break
+                for pattern in include_patterns:
                     if fnmatch(filtered_root, pattern):
-                        is_excluded = True
+                        is_excluded = False
                         break
+
                 if is_excluded:
                     continue
 
@@ -740,7 +797,7 @@ class Buildozer(object):
                 if fn.startswith('.'):
                     continue
 
-                # exclusion by pattern matching
+                # pattern matching
                 is_excluded = False
                 dfn = fn.lower()
                 if filtered_root:
@@ -749,6 +806,10 @@ class Buildozer(object):
                     if fnmatch(dfn, pattern):
                         is_excluded = True
                         break
+                for pattern in include_patterns:
+                    if fnmatch(dfn, pattern):
+                        is_excluded = False
+                        break
                 if is_excluded:
                     continue
 
@@ -790,9 +851,9 @@ class Buildozer(object):
             #self.error('Unable to patch main_py to add applibs directory.')
             return
 
-        header = ('import sys, os; '
-                  'sys.path = [os.path.join(os.getcwd(),'
-                  '"..", "_applibs")] + sys.path\n')
+        header = (b'import sys, os; '
+                   b'sys.path = [os.path.join(os.getcwd(),'
+                   b'"..", "_applibs")] + sys.path\n')
         with open(main_py, 'rb') as fd:
             data = fd.read()
         data = header + data
@@ -808,16 +869,35 @@ class Buildozer(object):
 
     @property
     def root_dir(self):
-        return realpath(join(dirname(self.specfilename)))
+        return realpath(dirname(self.specfilename))
+
+    @property
+    def user_build_dir(self):
+        """The user-provided build dir, if any."""
+        # Check for a user-provided build dir
+        # Check the (deprecated) builddir token, for backwards compatibility
+        build_dir = self.config.getdefault('buildozer', 'builddir', None)
+        if build_dir is not None:
+            # for backwards compatibility, append .buildozer to builddir
+            build_dir = join(build_dir, '.buildozer')
+        build_dir = self.config.getdefault('buildozer', 'build_dir', build_dir)
+
+        if build_dir is not None:
+            build_dir = realpath(join(self.root_dir, build_dir))
+
+        return build_dir
 
     @property
     def buildozer_dir(self):
-        if self.builddir:
-            return join(self.builddir, '.buildozer')
+        '''The directory in which to run the app build.'''
+        if self.user_build_dir is not None:
+            return self.user_build_dir
         return join(self.root_dir, '.buildozer')
 
     @property
     def bin_dir(self):
+        if self.user_bin_dir:
+            return self.user_bin_dir
         return join(self.root_dir, 'bin')
 
     @property
@@ -860,7 +940,6 @@ class Buildozer(object):
             return package_name
         return '{}.{}'.format(package_domain, package_name)
 
-
     #
     # command line invocation
     #
@@ -890,7 +969,10 @@ class Buildozer(object):
         print('Available targets:')
         targets = list(self.targets())
         for target, m in targets:
-            doc = m.__doc__.strip().splitlines()[0].strip()
+            try:
+                doc = m.__doc__.strip().splitlines()[0].strip()
+            except Exception:
+                doc = '<no description>'
             print('  {0:<18} {1}'.format(target, doc))
 
         print('')
@@ -926,7 +1008,8 @@ class Buildozer(object):
             for command, doc in commands:
                 if not doc:
                     continue
-                doc = doc.strip().splitlines()[0].strip()
+                doc = textwrap.fill(textwrap.dedent(doc).strip(), 59,
+                                    subsequent_indent=' ' * 21)
                 print('  {0:<18} {1}'.format(command, doc))
 
         print('')
@@ -1032,13 +1115,30 @@ class Buildozer(object):
                 return
             rmtree(self.global_buildozer_dir)
 
+    def cmd_appclean(self, *args):
+        '''Clean the .buildozer folder in the app directory.
+
+        This command specifically refuses to delete files in a
+        user-specified build directory, to avoid accidentally deleting
+        more than the user intends.
+        '''
+        if self.user_build_dir is not None:
+            self.error(
+                ('Failed: build_dir is specified as {} in the buildozer config. `appclean` will '
+                 'not attempt to delete files in a user-specified build directory.').format(self.user_build_dir))
+        elif exists(self.buildozer_dir):
+            self.info('Deleting {}'.format(self.buildozer_dir))
+            rmtree(self.buildozer_dir)
+        else:
+            self.error('{} already deleted, skipping.'.format(self.buildozer_dir))
+
     def cmd_help(self, *args):
         '''Show the Buildozer help.
         '''
         self.usage()
 
     def cmd_setdefault(self, *args):
-        '''Set the default command to do when to arguments are given
+        '''Set the default command to run when no arguments are given
         '''
         self.check_build_layout()
         self.state['buildozer:defaultcommand'] = args
@@ -1097,15 +1197,13 @@ class Buildozer(object):
                         value, section_base, profile))
                 self.config.set(section_base, name, value)
 
-
-
     def _get_config_list_values(self, *args, **kwargs):
         kwargs['with_values'] = True
         return self._get_config_list(*args, **kwargs)
 
     def _get_config_list(self, section, token, default=None, with_values=False):
         # monkey-patch method for ConfigParser
-        # get a key as a list of string, seperated from the comma
+        # get a key as a list of string, separated from the comma
 
         # check if an env var exists that should replace the file config
         set_config_token_from_env(section, token, self.config)
@@ -1155,6 +1253,15 @@ class Buildozer(object):
             return default
         return self.config.getboolean(section, token)
 
+    def _get_config_raw_default(self, section, token, default=None, section_sep="=", split_char=" "):
+        l_section = '{}:{}'.format(section, token)
+        if self.config.has_section(l_section):
+            return [section_sep.join(item) for item in self.config.items(l_section)]
+        if not self.config.has_option(section, token):
+            return default.split(split_char)
+        return self.config.get(section, token).split(split_char)
+
+
 def set_config_from_envs(config):
     '''Takes a ConfigParser, and checks every section/token for an
     environment variable of the form SECTION_TOKEN, with any dots
