@@ -14,9 +14,10 @@ def main():
     except BuildozerCommandException:
         # don't show the exception in the command line. The log already show
         # the command failed.
-        pass
+        sys.exit(1)
     except BuildozerException as error:
         Buildozer().error('%s' % error)
+        sys.exit(1)
 
 if __name__ == '__main__':
     main()
