@@ -1,5 +1,5 @@
 '''
-iOS target, based on kivy-ios project. (not working yet.)
+iOS target, based on kivy-ios project
 '''
 
 import sys
@@ -7,11 +7,12 @@ if sys.platform != 'darwin':
     raise NotImplementedError('Windows platform not yet working for Android')
 
 import plistlib
-from buildozer import BuildozerCommandException
+from buildozer import BuildozerCommandException, IS_PY3
 from buildozer.target import Target, no_config
 from os.path import join, basename, expanduser, realpath
 from getpass import getpass
 
+
 PHP_TEMPLATE = '''
 <?php
 // credits goes to http://jeffreysambells.com/2010/06/22/ios-wireless-app-distribution
@@ -20,7 +21,7 @@ $ipas = glob('*.ipa');
 $provisioningProfiles = glob('*.mobileprovision');
 $plists = glob('*.plist');
 
-$sr = stristr( $_SERVER['SCRIPT_URI'], '.php' ) === false ? 
+$sr = stristr( $_SERVER['SCRIPT_URI'], '.php' ) === false ?
     $_SERVER['SCRIPT_URI'] : dirname($_SERVER['SCRIPT_URI']) . '/';
 $provisioningProfile = $sr . $provisioningProfiles[0];
 $ipa = $sr . $ipas[0];
@@ -28,8 +29,8 @@ $itmsUrl = urlencode( $sr . 'index.php?plist=' . str_replace( '.plist', '', $pli
 
 
 if ($_GET['plist']) {
-    $plist = file_get_contents( dirname(__FILE__) 
-        . DIRECTORY_SEPARATOR 
+    $plist = file_get_contents( dirname(__FILE__)
+        . DIRECTORY_SEPARATOR
         . preg_replace( '/![A-Za-z0-9-_]/i', '', $_GET['plist']) . '.plist' );
     $plist = str_replace('_URL_', $ipa, $plist);
     header('content-type: application/xml');
@@ -59,6 +60,7 @@ li { padding: 1em; }
 '''
 
 class TargetIos(Target):
+    targetname = "ios"
 
     def check_requirements(self):
         checkbin = self.buildozer.checkbin
@@ -90,20 +92,11 @@ class TargetIos(Target):
         self.buildozer.debug(' -> found {0}'.format(xcode))
 
     def install_platform(self):
-        cmd = self.buildozer.cmd
-        self.ios_dir = ios_dir = join(self.buildozer.platform_dir, 'kivy-ios')
-        if not self.buildozer.file_exists(ios_dir):
-            cmd('git clone https://github.com/kivy/kivy-ios',
-                    cwd=self.buildozer.platform_dir)
-        elif self.platform_update:
-            cmd('git clean -dxf', cwd=ios_dir)
-            cmd('git pull origin master', cwd=ios_dir)
-
-        self.ios_deploy_dir = ios_deploy_dir = join(self.buildozer.platform_dir,
-                'ios-deploy')
-        if not self.buildozer.file_exists(ios_deploy_dir):
-            cmd('git clone https://github.com/phonegap/ios-deploy',
-                    cwd=self.buildozer.platform_dir)
+        self.ios_dir = self.install_or_update_repo('kivy-ios', platform='ios')
+        self.ios_deploy_dir = self.install_or_update_repo('ios-deploy',
+                                                          platform='ios',
+                                                          branch='1.7.0',
+                                                          owner='phonegap')
 
     def get_available_packages(self):
         available_modules = self.buildozer.cmd(
@@ -169,15 +162,19 @@ class TargetIos(Target):
         app_name = self.buildozer.namify(self.buildozer.config.get('app',
             'package.name'))
 
+        ios_frameworks = self.buildozer.config.getlist('app', 'ios.frameworks', '')
+        frameworks_cmd = ''
+        for framework in ios_frameworks:
+            frameworks_cmd += '--add-framework={} '.format(framework)
+
         self.app_project_dir = join(self.ios_dir, '{0}-ios'.format(app_name.lower()))
         if not self.buildozer.file_exists(self.app_project_dir):
-            self.buildozer.cmd('./toolchain.py create {0} {1}'.format(
-                app_name, self.buildozer.app_dir),
-                cwd=self.ios_dir)
+            create_cmd = './toolchain.py create {0}{1} {2}'.format(frameworks_cmd, app_name,
+                                                                   self.buildozer.app_dir)
+            self.buildozer.cmd(create_cmd, cwd=self.ios_dir)
         else:
-            self.buildozer.cmd('./toolchain.py update {0}-ios'.format(
-                app_name),
-                cwd=self.ios_dir)
+            update_cmd = './toolchain.py update {0}{1}-ios'.format(frameworks_cmd, app_name)
+            self.buildozer.cmd(update_cmd, cwd=self.ios_dir)
 
         # fix the plist
         plist_fn = '{}-Info.plist'.format(app_name.lower())
@@ -190,18 +187,14 @@ class TargetIos(Target):
         plist['CFBundleVersion'] = '{}.{}'.format(version,
                 self.buildozer.build_id)
 
-        # add icon
-        icon = self._get_icon()
-        if icon:
-            plist['CFBundleIconFiles'] = [icon]
-            plist['CFBundleIcons'] = {'CFBundlePrimaryIcon': {
-                'UIPrerenderedIcon': False, 'CFBundleIconFiles': [icon]}}
+        # add icons
+        self._create_icons()
 
         # ok, write the modified plist.
         plistlib.writePlist(plist, plist_rfn)
 
         mode = 'Debug' if self.build_mode == 'debug' else 'Release'
-        self.buildozer.cmd('xcodebuild -configuration {} clean build'.format(mode),
+        self.buildozer.cmd('xcodebuild -configuration {} ENABLE_BITCODE=NO clean build'.format(mode),
                 cwd=self.app_project_dir)
         ios_app_dir = '{app_lower}-ios/build/{mode}-iphoneos/{app_lower}.app'.format(
                 app_lower=app_name.lower(), mode=mode)
@@ -216,16 +209,42 @@ class TargetIos(Target):
         elif ioscodesign[0] not in ('"', "'"):
             ioscodesign = '"{}"'.format(ioscodesign)
 
-        ipa = join(self.buildozer.bin_dir, '{}-{}.ipa'.format(
+        intermediate_dir = join(self.ios_dir, '{}-{}.intermediates'.format(app_name, version))
+        xcarchive = join(intermediate_dir, '{}-{}.xcarchive'.format(
             app_name, version))
+        ipa_name = '{}-{}.ipa'.format(app_name, version)
+        ipa_tmp = join(intermediate_dir, ipa_name)
+        ipa = join(self.buildozer.bin_dir, ipa_name)
+        build_dir = join(self.ios_dir, '{}-ios'.format(app_name.lower()))
+
+        self.buildozer.rmdir(intermediate_dir)
+
+        self.buildozer.info('Creating archive...')
         self.buildozer.cmd((
-            '/usr/bin/xcrun '
-            '-sdk iphoneos PackageApplication {ios_app_dir} '
-            '-o {ipa} --sign {ioscodesign} --embed '
-            '{ios_app_dir}/embedded.mobileprovision').format(
-                ioscodesign=ioscodesign, ios_app_dir=ios_app_dir,
-                mode=mode, ipa=ipa),
-                cwd=self.ios_dir)
+                '/usr/bin/xcodebuild'
+                ' -alltargets'
+                ' -configuration {mode}'
+                ' -scheme {scheme}'
+                ' -archivePath "{xcarchive}"'
+                ' archive'
+                ' ENABLE_BITCODE=NO'
+            ).format(mode=mode, xcarchive=xcarchive, scheme=app_name.lower()),
+            cwd=build_dir)
+
+        self.buildozer.info('Creating IPA...')
+        self.buildozer.cmd((
+                '/usr/bin/xcodebuild'
+                ' -exportArchive'
+                ' -exportFormat IPA'
+                ' -archivePath "{xcarchive}"'
+                ' -exportPath "{ipa}"'
+                ' CODE_SIGN_IDENTITY={ioscodesign}'
+                ' ENABLE_BITCODE=NO'
+            ).format(xcarchive=xcarchive, ipa=ipa_tmp, ioscodesign=ioscodesign),
+            cwd=build_dir)
+
+        self.buildozer.info('Moving IPA to bin...')
+        self.buildozer.file_rename(ipa_tmp, ipa)
 
         self.buildozer.info('iOS packaging done!')
         self.buildozer.info('IPA {0} available in the bin directory'.format(
@@ -274,8 +293,7 @@ class TargetIos(Target):
             debug_mode=debug_mode, app_dir=ios_app_dir),
             cwd=self.ios_dir, show_output=True)
 
-    def _get_icon(self):
-        # check the icon size, must be 72x72 or 144x144
+    def _create_icons(self):
         icon = self.buildozer.config.getdefault('app', 'icon.filename', '')
         if not icon:
             return
@@ -283,40 +301,10 @@ class TargetIos(Target):
         if not self.buildozer.file_exists(icon_fn):
             self.buildozer.error('Icon {} does not exists'.format(icon_fn))
             return
-        output = self.buildozer.cmd('file {}'.format(icon),
-                cwd=self.buildozer.app_dir, get_stdout=True)[0]
-        if not output:
-            self.buildozer.error('Unable to read icon {}'.format(icon_fn))
-            return
-        # output is something like: 
-        # "data/cancel.png: PNG image data, 50 x 50, 8-bit/color RGBA,
-        # non-interlaced"
-        info = output.splitlines()[0].split(',')
-        fmt = info[0].split(':')[-1].strip()
-        if fmt != 'PNG image data':
-            self.buildozer.error('Only PNG icon are accepted, {} invalid'.format(icon_fn))
-            return
-        size = [int(x.strip()) for x in info[1].split('x')]
-        if size != [72, 72] and size != [144, 144]:
-            # icon cannot be used like that, it need a resize.
-            self.buildozer.error('Invalid PNG size, must be 72x72 or 144x144. Resampling.')
-            nearest_size = 144
-            if size[0] < 144:
-                nearest_size = 72
 
-            icon_basename = 'icon-{}.png'.format(nearest_size)
-            self.buildozer.file_copy(icon_fn, join(self.app_project_dir,
-                icon_basename))
-            self.buildozer.cmd('sips -z {0} {0} {1}'.format(nearest_size,
-                icon_basename), cwd=self.app_project_dir)
-        else:
-            # icon ok, use it as it.
-            icon_basename = 'icon-{}.png'.format(size[0])
-            self.buildozer.file_copy(icon_fn, join(self.app_project_dir,
-                icon_basename))
-
-        icon_fn = join(self.app_project_dir, icon_basename)
-        return icon_fn
+        self.buildozer.cmd('./toolchain.py icon {} {}'.format(
+            self.app_project_dir, icon_fn),
+            cwd=self.ios_dir)
 
     def _create_index(self):
         # TODO
@@ -408,7 +396,12 @@ class TargetIos(Target):
 
         save = None
         while save is None:
-            q = raw_input('Do you want to save the password (Y/n): ')
+            if IS_PY3:
+                input_func = input
+            else:
+                input_func = raw_input
+
+            q = input_func('Do you want to save the password (Y/n): ')
             if q in ('', 'Y'):
                 save = True
             elif q == 'n':
@@ -418,7 +411,7 @@ class TargetIos(Target):
 
         if save:
             with open(password_file, 'wb') as fd:
-                fd.write(password)
+                fd.write(password.encode())
 
 def get_target(buildozer):
     return TargetIos(buildozer)
